#
#  nn/callbacks/schedule.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch
from .base import TrainingCallback

class ParameterScheduler(TrainingCallback):
    """
    The parameter scheduler is able to change the value of a variable
    over the course of the training. The scheduler modifies parameters
    **in-place**, hence, you must pass tensors to be modified and you must
    never pass them to the CPU.
    """

    # MARK: Initialization
    def __init__(self, parameter, schedule, after_batch=False):
        """
        Initalizes a new scheduler for the given parameter.

        Parameters:
        -----------
        - parameter: torch.Tensor
            The parameter which should be modified over the course of the
            training.
        - schedule: func (float, int) -> float
            Function which should update the parameter (given as first
            argument) based on itself and the current epoch (second argument).
            The function must return the updated parameter. The scheduler
            function is called after every epoch.
        - after_batch: bool, default: False
            Whether to call the scheduler after every batch instead of after
            every epoch. The schedule function is then passed as second
            parameter the current iteration (number of all batches) instead of
            the epoch.
        """
        self.parameter = parameter
        self.schedule = schedule
        self.exec_after_batch = after_batch
        self.epoch = None
        self.iterations = None

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        self.iterations = 0

    def before_epoch(self, current, num_iterations):
        self.epoch = current

    def after_batch(self):
        self.iterations += 1
        self._update(True)

    def after_epoch(self, metrics):
        self._update(False)

    def after_training(self):
        self.epoch = None
        self.iterations = None

    # MARK: Private Methods
    def _update(self, is_batch_update):
        if is_batch_update != self.exec_after_batch:
            return
        if self.exec_after_batch:
            update = self.schedule(self.parameter.item(), self.iterations)
        else:
            update = self.schedule(self.parameter.item(), self.epoch)
        self.parameter.set_(torch.tensor(update))



#
#  __init__.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/03/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .multiprocessing import *
from .nn import *
from .utils import *



#
#  nn/modules/test_lstm.py
#  bxtorch/tests
#
#  Created by Oliver Borchert on June 13, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import unittest
import time
import numpy as np
import torch
import torch.nn as nn
from bxtorch.nn.modules.lstm import _LSTMCell

class TestLSTM(unittest.TestCase):

    def test_lstm_cell(self):
        nn_lstm = nn.LSTMCell(2, 2, bias=False)
        xnn_lstm = _LSTMCell(2, 2, bias=False)

        for _ in range(50):
            x = torch.rand(2).view(1, -1)
            hidden = torch.rand(2).view(1, -1)
            state = torch.rand(2).view(1, -1)

            val = float(np.random.rand(1))
            for p in nn_lstm.parameters():
                nn.init.constant_(p, val)
            for p in xnn_lstm.parameters():
                nn.init.constant_(p, val)

            with torch.no_grad():
                out_true = nn_lstm(x, (hidden, state))
                out_ours = xnn_lstm(x, (hidden, state))

            self.assertTrue(
                (out_true[0].numpy() == out_ours[0].numpy()).all()
            )
            self.assertTrue(
                (out_true[1].numpy() == out_ours[1].numpy()).all()
            )
        

if __name__ == '__main__':
    unittest.main()


#
#  nn/modules/lstm.py
#  bxtorch
#
#  Created by Oliver Borchert on June 13, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import numpy as np
import torch
import torch.nn as nn
import torch.jit as jit
from typing import List, Tuple, Optional

class StackedLSTM(nn.Module):
    """
    The stacked LSTM is an extension to PyTorch's native LSTM allowing stacked
    LSTMs with different hidden dimensions being stacked.
    """

    batch_first: jit.Final[bool]

    # MARK: Initialization
    def __init__(self, input_size, hidden_sizes, bias=True, batch_first=False,
                 cudnn=True):
        """
        Initializes a new stacked LSTM according to the given parameters.

        Parameters:
        -----------
        - input_size: int
            The dimension of the sequence's elements.
        - hidden_sizes: list of int
            The dimensions of the stacked LSTM's layers.
        - bias: bool, default: True
            Whether to use biases in the LSTM.
        - batch_first: bool, default: False
            Whether the batch or the sequence can be found in the first
            dimension.
        - cudnn: bool, default: True
            Whether to use PyTorch's LSTM implementation which uses cuDNN on
            Nvidia GPUs. You usually don't want to change the default value,
            however, PyTorch's default implementation does not allow higher-
            order gradients of the LSTMCell as of version 1.1.0. When this
            value is set to False, we therefore use a (slower) implementation
            of a LSTM cell which allows higher-order gradients.
        """
        super().__init__()
        self.batch_first = batch_first
        self.stacked_cell = StackedLSTMCell(
            input_size, hidden_sizes, bias=bias, cudnn=cudnn
        )

    # MARK: Instance Methods
    def forward(self, 
                inputs: torch.Tensor,
                initial_states: Optional[
                    List[Tuple[torch.Tensor, torch.Tensor]]] = None, 
                return_sequence: bool = True):
        """
        Computes the forward pass through the stacked LSTM.

        Parameters:
        -----------
        - inputs: torch.FloatTensor [S, B, N]
            The inputs fed to the LSTM one after the other. Sequence length S,
            batch size B, and input size N. If ``batch_first`` is set to True,
            the first and second dimension should be swapped.
        - initial_states: list of tuple of 
                (torch.FloatTensor [H_i], torch.FloatTensor [H_i]),
                default: None
            The initial states for all LSTM layers. The length of the list must
            match the number of layers in the LSTM, the sizes of the states must
            match the hidden sizes of the LSTM layers. If None is given, the
            initial states are defaulted to all zeros.
        - return_sequence: bool, default: True
            Whether to return all outputs from the last LSTM layer or only the
            last one.

        Returns:
        --------
        - torch.FloatTensor [S, B, K] or torch.FloatTensor [B, K]
            Depending on whether sequences are returned, either all outputs
            or only the output from the last cell are returned. If the stacked
            LSTM was initialized with ``batch_first``, the first and second
            dimension are swapped when sequences are returned.
        """
        if self.batch_first:
            inputs = inputs.transpose(1, 0)

        sequence_length = inputs.size(0)

        # Initialize the state to empty vectors is needed for jit to properly
        # compile the function
        if initial_states is None:
            states = [(torch.empty(0), torch.empty(0))]
        else:
            states = initial_states

        # Iterate over sequence
        outputs = []
        for n in range(sequence_length):
            output, states = self.stacked_cell(
                inputs[n], None if states[0][0].size(0) == 0 else states
            )
            if return_sequence or n == sequence_length - 1:
                outputs.append(output)

        if return_sequence:
            result = torch.stack(outputs)
            if self.batch_first: 
                # set batch first, sequence length second
                result = result.transpose(1, 0)
            return result
        else:
            return outputs[0]


class StackedLSTMCell(nn.Module):
    """
    Actually, the StackedLSTMCell can easily be constructed from existing 
    modules, however, a bug in PyTorch's JIT compiler prevents implementing
    anything where a stacked LSTM is used within a loop (see the following 
    issue: https://github.com/pytorch/pytorch/issues/18143). Hence, this
    class provides a single time step for a stacked LSTM.
    """

    cells: jit.Final[int]
    num_stacked: jit.Final[int]

    # MARK: Initialization
    def __init__(self, input_size, hidden_sizes, bias=True, cudnn=True):
        """
        Initializes a new stacked LSTM cell.

        Parameters:
        -----------
        - input_size: int
            The dimension of the input variables.
        - hidden_sizes: list of int
            The hidden dimension of the stacked LSTMs.
        - bias: bool, default: True
            Whether to use a bias term for the LSTM implementation.
        - cudnn: bool, default: True
            Whether to not use cuDNN. In almost all cases, you don't want to
            set this value to false, however, you will need to change it if
            you want to compute higher-order derivatives of a network with
            a stacked LSTM cell.
        """
        super().__init__()

        self.num_stacked = len(hidden_sizes)

        cell_class = nn.LSTMCell if cudnn else _LSTMCell

        cells = []
        dims = zip([input_size] + hidden_sizes, hidden_sizes)
        for i, (in_dim, out_dim) in enumerate(dims):
            cells.append(cell_class(in_dim, out_dim, bias=bias))
        self.cells = nn.ModuleList(cells)
        
    # MARK: Instance Methods
    def forward(self,
                x: torch.Tensor,
                initial_states: Optional[
                    List[Tuple[torch.Tensor, torch.Tensor]]] = None):
        """
        Computes the new hidden states and cell states for each stacked cell.

        Parameters:
        -----------
        - x: torch.Tensor [B, N]
            The input with batch size B and dimension N.
        - states: list of tuple of (torch.Tensor [B, D], torch.Tensor [B, D]),
                default: None
            The states for each of the cells where each state is expected to
            have a size with batch size B and (respective) hidden dimension D.

        Returns:
        --------
        - torch.Tensor [B, D]
            The output, i.e. the hidden state of the deepest cell. Only given
            for convenience as it can be extracted from the other return value.
        - list of tuple of (torch.Tensor [B, D], torch.Tensor [B, D])
            The new hidden states and cell states for all cells.
        """
        if initial_states is None:
            # JIT Compatibility
            states = [
                (torch.empty(0), torch.empty(0))
                for _ in range(self.num_stacked)
            ]
        else:
            states = initial_states
        
        i = 0
        for cell in self.cells:
            x, next_cell = cell(
                x, None if states[i][0].size(0) == 0 else states[i]
            )
            states[i] = (x, next_cell)
            i += 1

        return x, states


class _LSTMCell(nn.Module):
    """
    LSTMCell which does not have cuDNN support but allows for higher-order
    gradients.
    Consult PyTorch's LSTMCell for documentation on the class's initialization
    parameters and how to call it.
    """

    hidden_size: jit.Final[int]
    has_bias: jit.Final[bool]

    # MARK: Initialization
    def __init__(self, input_size, hidden_size, bias=True):
        super().__init__()

        self.hidden_size = hidden_size

        self.input_weight = nn.Parameter(
            torch.FloatTensor(input_size, 4 * hidden_size)
        )
        self.hidden_weight = nn.Parameter(
            torch.FloatTensor(hidden_size, 4 * hidden_size)
        )

        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(4 * hidden_size))
            self.has_bias = True
        else:
            self.has_bias = False

        self.reset_parameters()

    # MARK: Instance Methods
    def reset_parameters(self):
        sqrt_hidden = np.sqrt(1 / self.hidden_size)
        init_from = (-sqrt_hidden, sqrt_hidden)
        for p in self.parameters():
            nn.init.uniform_(p, *init_from)

    def forward(self, 
                x_in: torch.Tensor,
                state: Optional[Tuple[torch.Tensor, torch.Tensor]] = None):

        if state is None:
            size = (x_in.size(0), self.hidden_size)
            hidden_state = torch.zeros(
                *size, dtype=torch.float, device=x_in.device
            )
            cell_state = torch.zeros(
                *size, dtype=torch.float, device=x_in.device
            )
        else:
            hidden_state, cell_state = state

        # 1) Perform matrix multiplications for input and last hidden state
        if self.has_bias:
            x = torch.addmm(self.bias, x_in, self.input_weight)
            h = torch.addmm(self.bias, hidden_state, self.hidden_weight)
        else:
            x = x_in.matmul(self.input_weight)
            h = hidden_state.matmul(self.hidden_weight)

        forget_gate_x, input_gate_x_1, input_gate_x_2, output_gate_x = \
            x.split(self.hidden_size, dim=1)
        forget_gate_h, input_gate_h_1, input_gate_h_2, output_gate_h = \
            h.split(self.hidden_size, dim=1)

        # 2) Forget gate
        forget_gate = torch.sigmoid(forget_gate_x + forget_gate_h)

        # 3) Input gate
        input_gate_1 = torch.sigmoid(input_gate_x_1 + input_gate_h_1)
        input_gate_2 = torch.tanh(input_gate_x_2 + input_gate_h_2)
        input_gate = forget_gate * cell_state + input_gate_1 * input_gate_2

        # 4) Output gate
        output_gate_1 = torch.sigmoid(output_gate_x + output_gate_h)
        output_gate = output_gate_1 * torch.tanh(input_gate)

        return output_gate, input_gate



#
#  utils/stdmp.py
#  bxtorch
#
#  Created by Oliver Borchert on June 28, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import time

def terminate(*processes, force=True):
    """
    Terminates all given processes. If a process does not finish within a
    second, it is terminated (signal 15).

    Parameters:
    -----------
    - processes: varargs
        The processes to terminate.
    - force: bool
        Whether to kill processes (with signal 9) in case they do terminate
        on signal 15 after 0.05 seconds.
    """
    deadline = time.time() + 1
    for p in processes:
        p.join(max(deadline - time.time(), 0.1))
        if p.is_alive():
            p.terminate()
            if force:
                time.sleep(0.05)
                if p.is_alive():
                    p.kill()



#
#  graph/test_numpy.py
#  bxtorch/tests
#
#  Created by Oliver Borchert on June 16, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import sys
sys.path.append('../../')

import unittest
import numpy as np
from bxtorch.utils import numpy as xnp

class TestGraphStats(unittest.TestCase):

    def test_find_contexts(self):
        # Without pivot
        seed = np.array([
            [1, 2, 3, 4],
            [5, 6, 7, 8]
        ])
        result = xnp.find_contexts(seed, 2)
        
        expected = np.array([
            [[1,2], [2,3], [3,4]],
            [[5,6], [6,7], [7,8]]
        ])
        self.assertTrue((expected == result).all())

        # With pivot
        result = xnp.find_contexts(seed, 3, exclude_pivot=True)

        expected_contexts = np.array([
            [[1,3], [2,4]],
            [[5,7], [6,8]]
        ])
        expected_targets = np.array([
            [2, 3],
            [6, 7]
        ])

        self.assertTrue((expected_contexts == result[0]).all())
        self.assertTrue((expected_targets == result[1]).all())

    def test_find_contexts_1d(self):
        seed = np.array([
            1, 2, 3, 4, 5, 6
        ])
        expected = np.array([
            [1, 2], [2, 3], [3, 4], [4, 5], [5, 6]
        ])
        result = xnp.find_contexts(seed, 2)

        self.assertTrue((expected == result).all())

    def test_find_contexts_nd(self):
        seed = np.array([
            [
                [1, 2, 3, 4],
                [5, 6, 7, 8]
            ],
            [
                [9, 10, 11, 12],
                [13, 14, 15, 16]
            ]
        ])
        expected = np.array([
            [
                [[1, 2, 3], [2, 3, 4]],
                [[5, 6, 7], [6, 7, 8]]
            ],
            [
                [[9, 10, 11], [10, 11, 12]],
                [[13, 14, 15], [14, 15, 16]]
            ]
        ])
        result = xnp.find_contexts(seed, 3)

        self.assertTrue((expected == result).all())


if __name__ == '__main__':
    unittest.main()



#
#  nn/modules/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import numpy as np
from sklearn.utils.class_weight import compute_class_weight
import torch
import torch.nn as nn

class BCELoss(nn.Module):
    """
    Binary cross entropy loss which allows for easy weighting of classes as
    well as an index which is ignored in calculations. With default parameters,
    this loss defaults to PyTorch's built-in BCELoss.
    """
    
    # MARK: Initialization
    def __init__(self, labels=None, ignore_index=None):
        """
        Initializes a new BCELoss.

        Parameters:
        -----------
        - labels: np.ndarray [N], default: None
            If given, class weights are computed according to the given labels.
        - ignore_index: int, default: None
            If given, this index in the labels is ingored when computing the
            loss.
        """
        super().__init__()
        
        self.ignore_index = ignore_index

        if labels is not None:
            weights = compute_class_weight(
                'balanced', np.arange(np.max(labels) + 1), labels
            )
            self.weights = torch.as_tensor(weights, dtype=torch.float)
            self.loss = nn.BCELoss(reduction='none')
        else:
            self.weights = None
            self.loss = nn.BCELoss()

    # MARK: Instance Methods
    def forward(self, y_pred, y_true):
        if self.ignore_index is not None:
            mask = ~(y_true == self.ignore_index)
            y_pred = y_pred[mask]
            y_true = y_true[mask]

        if self.weights is not None:
            self.class_weights = self.class_weights.to(y_pred.device)
            bce = self.loss(y_pred, y_true)
            return (bce * self.class_weights[y_true.long()]).mean()

        return self.loss(y_pred, y_true)



#
#  wrappers.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/04/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import json
import numpy as np
from bxtorch.utils.stdlib import flatten
try:
    import matplotlib.pyplot as plt
except:
    pass

class History:
    """
    This class summarizes metrics obtained during training of a model.
    """

    # MARK: Static Methods
    @staticmethod
    def load(file):
        """
        Loads the history object from the specified file.

        Parameters:
        -----------
        - file: str
            The JSON file to load the history object from.
        """
        with open(file, 'r') as f:
            loaded = json.load(f)
        return History(**loaded)

    @staticmethod
    def cat(hist1, hist2):
        """
        Concatenates the two history objects.

        Parameters:
        -----------
        - hist1: bxtorch.nn.History
            The first history.
        - hist2: bxtorch.nn.History
            The second history.

        Returns:
        --------
        - bxtorch.nn.History
            The concatenated history object.
        """
        return History(
            hist1.duration + hist2.duration,
            hist1._metrics + hist2._metrics
        )
    
    # MARK: Initialization
    def __init__(self, duration, metrics):
        """
        Initializes a new history object.

        Parameters:
        -----------
        - metrics: list of dict
            The metrics from each training step.
        """
        self.duration = duration
        self._metrics = metrics

    # MARK: Instance Methods
    @property
    def keys(self):
        """
        Returns the available keys that this history object provides.

        Returns:
        --------
        - list of str
            The keys.
        """
        return sorted(self._metrics[0].keys())

    def save(self, file):
        """
        Saves the history object to the specified file as JSON.

        Parameters:
        -----------
        - file: str
            The file (with .json extension) to save the history to.
        """
        obj = {'duration': self.duration, 'metrics': self._metrics}
        with open(file, 'w+') as f:
            json.dump(obj, f, indent=4, sort_keys=True)

    # MARK: Special Methods
    def __len__(self):
        return len(self._metrics)

    def __getattr__(self, name):
        if name == '_metrics':
            raise AttributeError()
        if len(self._metrics) == 0:
            raise AttributeError(
                f'Length of history is 0. Metric {name} cannot be accessed.'
            )
        if name in self._metrics[0]:
            result = [m[name] for m in self._metrics]
            if name == 'micro_train_loss':
                return flatten(result)
            return result
        else:
            raise AttributeError(f'Metric {name} does not exist.')
            

class Evaluation:
    """
    This class summarizes metrics obtained when evaluating a model.
    """

    # MARK: Static Methods
    @staticmethod
    def merge(eval1, eval2, *eval_other):
        """
        Merges multiple evaluations into one.

        Parameters:
        -----------
        - eval1: torch.nn.training.wrappers.Evaluation
            The first evaluation.
        - eval2: torch.nn.training.wrappers.Evaluation
            The second evaluation.
        - eval_other: varargs of torch.nn.training.wrappers.Evaluation
            Additional evaluations.

        Returns:
        --------
        - torch.nn.training.wrappers.Evaluation
            The merged evaluation.
        """
        result = Evaluation({})
        metrics = {}
        for e in [eval1, eval2] + list(eval_other):
            for k, v in e._metrics.items():
                metrics[k] = v
        result._metrics = metrics
        return result

    # MARK: Initialization
    def __init__(self, metrics, weights=None):
        """
        Initializes a new evaluation wrapper from the given metrics.

        Parameters:
        -----------
        - metrics: dict of str -> list of float
            The metrics computed. The keys define the metric names, the values
            provide the metrics computed over all batches.
        - weights: list of int, default: None
            The length of the batches. Should be supplied if the batches for
            which the metrics have been computed have differing sizes. If so,
            this list must contain the sizes of all batches.
        """
        result = {}
        for metric, values in metrics.items():
            if weights is None or len(weights) == 0:
                average = np.mean(values)
            else:
                weights = np.array(weights)
                values = np.array(values)
                average = weights.dot(values) / np.sum(weights)
            result[metric] = average
        self._metrics = result

    # MARK: Instance Methods
    @property
    def keys(self):
        """
        Returns the available keys for this evaluation object.

        Returns:
        --------
        - list of str
            The keys.
        """
        return sorted(self._metrics.keys())

    def to_dict(self):
        """
        Returns the evaluation object as plain dictionary.

        Returns:
        --------
        - dict
            The dictionary.
        """
        return {k: v for k,v in self._metrics.items() if not k.startswith('_')}

    def with_prefix(self, prefix):
        """
        Initializes a new evaluation wrapper where all metrics are prefixed
        with the given prefix. Useful when merging evaluations.

        Parameters:
        -----------
        - prefix: str
            The prefix to use.

        Returns:
        --------
        - torch.nn.training.wrappers.Evaluation
            The resulting evaluation.
        """
        result = Evaluation({})
        result._metrics = {f'{prefix}{k}': v for k, v in self._metrics.items()}
        return result

    # MARK: Special Methods
    def __contains__(self, item):
        return item in self._metrics

    def __getitem__(self, item):
        return self._metrics[item]

    def __getattr__(self, name):
        if name == '_metrics':
            raise AttributeError()
        try:
            return self._metrics[name]
        except KeyError:
            class_name = self.__class__.__name__
            raise AttributeError(
                f'{class_name} does not provide metric {name}.'
            )



#
#  nn/utils/model.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch.nn as nn

class DataParallel(nn.DataParallel):
    """
    Replacement for PyTorch's default ``nn.DataParallel`` class. It enables
    accessing the attributes of the underlying module. If the attribute being
    accessed is a module itself, it is also returned being wrapped as
    DataParallel. Call ``.module`` to access the raw model.
    """

    # MARK: Initialization
    def __init__(self, model, device_ids=None, output_device=None, dim=0):
        super(DataParallel, self).__init__(
            model, device_ids, output_device, dim
        )
        self._cache = {}

    # MARK: Special Methods
    def __getattr__(self, name):
        try:
            return super(DataParallel, self).__getattr__(name)
        except AttributeError:
            value = getattr(self.module, name)
            if isinstance(value, nn.Module):
                if value not in self._cache:
                    self._cache[value] = DataParallel(
                        value, device_ids=self.device_ids,
                        output_device=self.output_device,
                        dim=self.dim
                    )
                return self._cache[value]
            return value



#
#  optim/lamb.py
#  bxtorch
#
#  Created by Oliver Borchert on June 16, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import math
import torch
import torch.optim as optim

class LAMB(optim.Optimizer):
    """
    Optimizer presented in "Large Batch Optimization for Deep Learning: Training
    Bert in 76 Minutes" (You et al., 2019).

    The LAMB optimizer ("Layer-wise Adaptive Moments optimizer for Batch 
    training") enables training on very large batches and provides an alternative for Adam whose performance deteriorates for large batches.
    """

    # MARK: Initialization
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, lr_decay=0):
        """
        Initializes a new LAMB optimizer.

        Parameters:
        -----------
        - params: iterable of torch.Tensor or dict of str -> torch.Tensor
            The parameters to optimize, optionally overriding default values
            for parameter groups.
        - lr: float, default: 1e-3
            The learning rate.
        - betas: tuple of (float, float), default: (0.9, 0.999)
            The betas used to compute the running average of gradients.
        - eps: float, default: 1e-8
            Epsilon parameter for numerical stability.
        - weight_decay: float, default: 0
            L2 penalty to apply.
        - lr_decay: float, default: 0
            Learning rate decay over each update.
        """
        assert lr > 0, "Learning rate must be greater 0."
        assert betas[0] > 0 and betas[1] > 0, "Beta values must be greater 0."
        assert betas[0] < 1 and betas[1] < 1, "Beta values must be smaller 1."
        assert eps > 0, "Epsilon must be gerater 0."

        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
            lr_decay=lr_decay
        )
        super().__init__(params, defaults)

    def step(self, closure=None):
        """
        Performs a single optimization step.

        Parameters:
        -----------
        - closure: callable, default: None
            A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data

                if grad.is_sparse:
                    raise RuntimeError(
                        "LAMB does not support sparse gradients. " + 
                        "Consider using SparseAdam instead."
                    )

                state = self.state[p]

                # Initialize state if needed
                if len(state) == 0: 
                    state['step'] = 0
                    state['exp_avg'] = torch.zeros_like(p.data)
                    state['exp_avg_sq'] = torch.zeros_like(p.data)

                # Get relevant parameters
                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                # Next step
                state['step'] += 1

                # Update learning rate if needed
                if group['lr_decay'] > 0:
                    decay = group['lr_decay']
                    lr = group['lr'] * (1 / 1 + decay * state['step'])
                else:
                    lr = group['lr']

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                step_size = lr * math.sqrt(bias_correction2) / bias_correction1
                adam_step = exp_avg / denom

                # L2 penalty
                if group['weight_decay'] != 0:
                    adam_step.add_(group['weight_decay'], p.data)

                # Compute trust ratio and cap at 10
                r1 = p.data.norm()
                r2 = adam_step.norm()
                trust_ratio = min(r1 / r2, 10) if r1 != 0 and r2 != 0 else 1

                # Update weights
                p.data.add_(-step_size * trust_ratio, adam_step)

        return loss
        


#
#  utils/numpy.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
import numba

def find_contexts(array, context_size, exclude_pivot=False):
    """
    Given an array of size [N], the function extracts subarrays of size C
    (where C is ``context_size`), hence yielding an array of size [W, C] where 
    W is the number of subarrays, given as ``N - C + 1``.

    If the given array has multiple dimensions, the subarrays are only
    extracted for the last dimension.

    Parameters:
    -----------
    - array: numpy.ndarray
        The array to extract subarrays from.
    - context_size: int
        The size of the subarrays to extract.
    - exclude_pivot: bool, default: False
        If this option is set, the number of subarrays returned does not change,
        however, the size of the subarrays reduces to C - 1 and the middle value
        is not included. Instead, a second value is returned, containing all
        these middle values. Setting this option is useful in some settings
        such as sampling windows from random walks for DeepWalk. Note that the
        context size must be an odd number if this option is set.

    Returns:
    --------
    - numpy.ndarray
        The subarrays whose dimension is one more than the given array and where
        all dimensions (except for the last one) are preserved.
    - numpy.ndarray
        The target values if the ``exclude_pivot`` option is set. This array
        has the same dimension as the given array, however, its last dimension
        is equal to the number of contexts found in this dimension.
    """
    assert context_size > 1, \
        "The context size must be at least 2."
    assert not exclude_pivot or context_size % 2 == 1, \
        "If the exclude_pivot option is set, the context size must be odd."
    
    old_shape = array.shape
    array = array.reshape((-1, array.shape[-1]))
    if not exclude_pivot:
        result = _find_contexts_without_pivot(array, context_size)
        return result.reshape(old_shape[:-1] + result.shape[1:])
    else:
        contexts, targets = _find_contexts_with_pivot(array, context_size)
        return (
            contexts.reshape(old_shape[:-1] + contexts.shape[1:]),
            targets.reshape(old_shape[:-1] + targets.shape[-1:])
        )


@numba.njit
def _find_contexts_without_pivot(array, context_size):
    # Dimension of array is 2
    contexts = np.empty(
        (array.shape[0], array.shape[1] - context_size + 1, context_size),
        dtype=array.dtype
    )

    for i in range(array.shape[0]):
        sliding = []
        for j, item in enumerate(array[i]):
            sliding.append(item)
            if j < context_size - 1:
                continue
            else:
                contexts[i][j - context_size + 1] = sliding
                sliding = sliding[1:]

    return contexts


@numba.njit
def _find_contexts_with_pivot(array, context_size):
    contexts = np.empty(
        (array.shape[0], array.shape[1] - context_size + 1, context_size - 1)
    )
    targets = np.empty(contexts.shape[:2])

    halfc = (context_size - 1) // 2

    for i in range(array.shape[0]):
        sliding = []
        for j, item in enumerate(array[i]):
            sliding.append(item)
            if j < context_size - 1:
                continue
            else:
                contexts[i][j - context_size + 1] = \
                    sliding[:halfc] + sliding[halfc+1:]
                targets[i][j - context_size + 1] = sliding[halfc]
                sliding = sliding[1:]

    return contexts, targets



#
#  __init__.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/05/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .base import BaseEngine
from .supervised import SupervisedEngine
from .wrappers import History, Evaluation



#
#  nn/training/proxy.py
#  bxtorch
#
#  Created by Oliver Borchert on June 18, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch.optim as optim
from bxtorch.utils.stdlib import flatten

class ProxyOptimizer(optim.Optimizer):
    """
    Optimizer for use with Hogwild!. It is only intended for internal usage,
    do not use this class directly. The idea is simply to provide a proxy for 
    an optimizer on a shared model while gradients are collected from another
    model (which is e.g. optimized on the GPU). It should generally be used
    with some appropriate callbacks during training.
    """

    # MARK: Initialization
    def __init__(self, optimizer, shared_optimizer):
        """
        Initializes a new proxy optimizer. Note that two optimizers are 
        required (instead of a model and an optimizer) as an optimizer must
        not necessarily optimize all parameters of a model. Evidently, both
        given optimizers must optimize the same parameters (of different
        models with the same architecture).

        Parameters:
        -----------
        - optimizer: torch.optim.Optimizer
            The "actual" optimizer.
        - shared_optimizer: torch.optim.Optimizer
            The optimizer to which all calls are proxied. 
        """
        # Initialize super with all parameters such that operations like
        # ``zero_grad`` still work as expected. Otherwise, do not pass any
        # configuration information.
        super().__init__(
            flatten([g['params'] for g in optimizer.param_groups]), {}
        )

        self.optimizer = optimizer
        self.shared_optimizer = shared_optimizer
    
    # MARK: Instance Methods
    def step(self, closure=None):
        """
        Performs an optimization step by passing the gradients of the model
        to the shared optimizer and performing an optimization step on the
        shared model.

        Parameters:
        -----------
        - closure: callable
            A closure to reevaluate the model and return the loss.
        """
        # 1) Move the gradients to the shared model

        # Optimizers must have the same param groups, otherwise behavior is
        # undefined
        param_groups = zip(
            self.optimizer.param_groups, 
            self.shared_optimizer.param_groups
        )
        for parameter_group, shared_parameter_group in param_groups:
            parameters = zip(
                parameter_group['params'],
                shared_parameter_group['params']
            )
            # Now, actually move parameters, try not to block
            for parameter, shared_parameter in parameters:
                device = shared_parameter.device
                grad = parameter.grad
                # We set the gradient directly as we then don't need to zero
                # the gradient
                shared_parameter._grad = grad.to(device, non_blocking=True)

        # 2) Perform update step on shared model
        # The model weights from the updated model should then be loaded by
        # an appropriate callback.
        return self.shared_optimizer.step(closure)



#
#  nn/functional/gumbel.py
#  bxtorch
#
#  Created by Oliver Borchert on June 12, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import torch
import torch.jit as jit
import torch.nn as nn
import torch.nn.functional as F

@jit.script
def gumbel_softmax(logits: torch.Tensor,
                   tau: float = 1, 
                   hard: bool = False,
                   eps: float = 1e-10, 
                   dim: int = -1):
    """
    Numerically stable version of PyTorch's builtin gumbel softmax.

    Parameters:
    -----------
    - logits: torch.Tensor
        The values fed into the gumbel softmax.
    - tau: float, default: 1
        Temperature parameter for the Gumbel distribution.
    - hard: bool, default: False
        Whether to obtain a one-hot output.
    - eps: float, default: 1e-10
        Constant for numerical stability.
    - dim: int, default: -1
        The dimension over which to apply the softmax.
    """
    uniform_sample = torch.empty_like(logits).uniform_(0.0, 1.0)
    gumbel_sample = -torch.log(eps - torch.log(uniform_sample + eps))

    y_soft = F.softmax((logits + gumbel_sample) / tau, dim=dim)

    if hard:
        index = y_soft.max(dim, keepdim=True)[1]
        y_hard = torch.zeros_like(logits).scatter_(dim, index, 1.0)
        return y_hard - y_soft.detach() + y_soft
    else:
        return y_soft



#
#  nn/utils/config.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import copy
from abc import ABC, abstractclassmethod
from collections import OrderedDict
import json
import torch
import torch.nn as nn
from bxtorch.utils.stdio import ensure_valid_directories

class Config(ABC):
    """
    Base class which standardizes configuration for PyTorch modules. For a
    specific configuration subclass, the configuration values are automatically
    deduced from the class's annotations. All annotations which do not define
    a default value are considered required.
    """

    # MARK: Static Methods
    @classmethod
    def load(cls, file):
        """
        Loads a configuration from the specified JSON file.

        Parameters:
        -----------
        - file: str
            The full path to the file. The extension does not need to be
            specified, but must be .json.

        Returns:
        --------
        - bxtorch.nn.Config
            The loaded configuration.
        """
        if not file.endswith('.json'):
            file += '.json'
        with open(file, 'r') as f:
            return cls(**json.load(f))

    @classmethod
    def _get_fields(cls):
        """
        Aggregates all available fields and their types from the class itself
        and its superclass (assuming it is a configuration as well).

        Returns:
        --------
        - dict
            Mapping from field names to types.
        """
        annotations = {}
        if hasattr(super(), '_get_fields'):
            annotations = {**annotations, **super()._get_fields()}
        if hasattr(cls, '__annotations__'):
            annotations = {**annotations, **cls.__annotations__}
        return annotations
    
    # MARK: Initialization
    def __init__(self, **kwargs):
        """
        Initializes a new config from the given parameters.
        """
        # 1) Get all available configuration options
        all_fields = type(self)._get_fields()
        keys = set(all_fields.keys())
        params = {}

        # 2) Set all parameters that are passed via the initializer
        for k, v in kwargs.items():
            if not k in keys:
                raise ValueError(
                    f'Unknown configuration parameter {k}.'
                )
            params[k] = v
            keys.remove(k)

        # 3) For all keys that are not set manually, assign default if available
        for k in keys:
            if not hasattr(type(self), k):
                raise ValueError(
                    f'Missing required configuration parameter {k}.'
                )
            params[k] = getattr(type(self), k)

        # 4) Check for all parameters if their types are correct
        for k, v in params.items():
            if not isinstance(v, all_fields[k]):
                raise ValueError(
                    f'Parameter {k} should have type {all_fields[k]} but '
                    f'has type {type(v)}.'
                )
        
        self._params = params

        # 5) Finally, set parameters on self and validate
        for k, v in params.items():
            setattr(self, k, v)

        if not self.is_valid():
            raise ValueError(
                f'Invalid configuration parameters. Check the documentation '
                f'for valid options.'
            )

    # MARK: Instance Methods
    def save(self, file):
        """
        Saves the configuration to a JSON file.

        Parameters:
        -----------
        - file: str
            The full path to the file. The extension does not need to be
            specified, but must be .json.
        """
        ensure_valid_directories(file)
        if not file.endswith('.json'):
            file += '.json'
        with open(file, 'w+') as f:
            json.dump(self._params, f, indent=4, sort_keys=True)

    def is_valid(self):
        """
        Checks whether the given configuration is valid. This method should be
        overridden if required. The default implementation implies validity for
        all inputs matching the specified types.

        Returns:
        --------
        - bool
            Whether the configuration is valid.
        """
        return True

    # MARK: Special Methods
    def __repr__(self):
        return self._params.__repr__()


class Configurable(ABC):
    """
    Mixin class to make torch.nn.Module configurable in an easy and standardized
    way. For initialization, a module is then given a single config file of 
    type ``bxtorch.nn.Config``. Properties of the configuration can be easily 
    accessed via ``self.<property>``. Additionally, this class makes it easy
    to save and load the module in a consistent way. The binding to the
    configuration class is achieved by setting the __

    When subclassing a PyTorch module, make sure to include the Configurable
    mixin as *first* dependency.

    Note:
    -----
    Only top-level modules should use this mixin to profit from saving/loading
    modules easily.
    """

    # MARK: Static Methods
    @classmethod
    def load(cls, file):
        """
        Loads the model by loading a configuration defining the architecture
        and another file containing the weights for this architecture.

        Parameters:
        -----------
        - file: str
            The full path to the file. The configuration will be loaded from
            <file>.json and the model's weights will be loaded from <file>.pt.

        Returns:
        --------
        - bxtorch.nn.Configurable
            The loaded model.
        """
        if hasattr(cls, '__config__'):
            config = cls.__config__.load(file)
            model = cls(config)
        else:
            model = cls(None)
        params = torch.load(f'{file}.pt')
        model.load_state_dict(params)
        return model

    # MARK: Initialization
    def __init__(self, *args, **kwargs):
        """
        Initializes the configurable model. You can either pass an instance of
        the model's configuration class or the keyword arguments which you
        would normally pass to initialize the configuration instance. The
        configuration is then created implicitly.

        Parameters:
        -----------
        - args: variadic arguments
            Either zero or one parameters. If one is given, it must be an
            instance of the model's configuration class.
        - kwargs: keyword arguments
            The arguments passed to the initializer of the configuration class.
            Must only be given if the configuration instance is not defined.
        """
        super(Configurable, self).__init__()

        assert len(args) <= 1, \
            "The only variadic argument must be a configuration instance."

        if len(args) == 1:
            assert len(kwargs) == 0, \
                "Variadic arguments must not be given if a another " + \
                "argument is defined."
            assert args[0] is None or \
                isinstance(args[0], type(self).__config__), \
                "Given configuration does not match the model's config class."
            
            self._config = args[0]
        else:
            self._config = type(self).__config__(**kwargs)

    # MARK: Computed Properties
    @property
    def num_parameters(self):
        """
        Returns the number of parameters that this model contains.

        Returns:
        --------
        - int
            The number of parameters.
        """
        return sum(p.numel() for p in self.parameters())

    @property
    def num_trainable_parameters(self):
        """
        Returns the number of trainable parameters that this model contains.

        Returns:
        --------
        - int
            The number of trainable parameters.
        """
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    # MARK: Instance Methods
    def save(self, file):
        """
        Saves the model and its configuration to two files.

        Parameters:
        -----------
        - file: str
            The full path to the file. The configuration will be saved to
            <file>.json while the model's weights will be saved to <file>.pt.
        """
        ensure_valid_directories(file)
        self._config.save(file)
        # Ensure saving CPU model
        state_dict = copy.deepcopy(self.state_dict())
        result = OrderedDict()
        for k, v in state_dict.items():
            result[k] = v.cpu()
        result._metadata = state_dict._metadata
        torch.save(result, f'{file}.pt')

    # MARK: Special Methods
    def __getattr__(self, name):
        try:
            return super(Configurable, self).__getattr__(name)
        except AttributeError:
            if name == '_config':
                raise AttributeError()
            return getattr(self._config, name)



#
#  nn/functional/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .gumbel import gumbel_softmax
from .metrics import accuracy, precision, average_precision, recall, \
    roc_auc_score
from .random import generate_noise



#
#  extensions.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/05/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import math
import numpy as np
import torch.utils.data as data

###############
### METHODS ###
###############
def loader(self, **kwargs):
    """
    Returns a data loader for this dataset. If the dataset defines a 
    `collate_fn` function, this is automatically set 

    Parameters:
    -----------
    - kwargs: keyword arguments
        Paramaters passed directly to the DataLoader.

    Returns:
    --------
    - torch.utils.data.DataLoader
        The data loader with the specified attributes.
    """
    if hasattr(self, 'collate_fn'):
        kwargs['collate_fn'] = self.collate_fn
    return data.DataLoader(self, **kwargs)


def random_split(self, *sizes, seed=None):
        """
        Splits the dataset randomly into multiple subsets.

        Parameters:
        -----------
        - sizes: variadic argument of float
            The sizes of the splits, given as fraction of the size of the
            dataset. Hence, the sizes must sum to 1.
        - seed: int, default: None
            If given, uses the specified seed to sample the indices for each
            subset.

        Returns:
        --------
        - list of bxtorch.nn.Subset
            The random splits of this dataset.
        """
        assert math.isclose(sum(sizes), 1), \
            "Sizes do not sum to 1."

        randomizer = np.random.RandomState(seed)
        
        # Get subset sizes
        nums = []
        for i, size in enumerate(sizes):
            if i == len(sizes) - 1:
                nums.append(len(self) - sum(nums))
            else:
                nums.append(int(np.round(size * len(self))))

        # Get subset indices
        indices = randomizer.permutation(len(self))
        index_choices = []
        c = 0
        for num in nums:
            index_choices.append(indices[c:c+num])
            c += num

        # Generate subsets
        if hasattr(self, 'collate_fn'):
            # Dynamically define a new subset
            subset_cls = type(
                '_Subset', (data.Subset,), {'collate_fn': self.collate_fn}
            )
        else:
            subset_cls = data.Subset
        
        return [
            subset_cls(self, indices) for indices in index_choices
        ]

##################
### EXTENSIONS ###
##################
data.Dataset.loader = loader
data.Dataset.random_split = random_split



#
#  supervised.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/04/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import copy
from .base import BaseEngine

class SupervisedEngine(BaseEngine):
    """
    Trainer to be used in a supervised learning setting.

    A dataset supplied to the `train` or `evaluate` method must supply tuples 
    with the following items:

    * x: input to the model
    * y: correct output

    For the `predict method, the data must be supplied as follows:

    * x: input to the model

    Further, the `train` method requires the following parameters:

    - optimizer: torch.optim.Optimizer or str
        The optimizer to use for optimizing the model's weights. If a string is 
        given, the corresponding optimizer is used with default 
        hyperparameters. Be aware that this optimizer is kept in memory as long 
        as the trainer exists.
    - loss: func (torch.Tensor, torch.Tensor) -> float
        The loss function to use.
    - gradient_accumulation_steps: int, default: 1
        The number of batches which should be used for a single update step.
        Gradient accumulation can be useful if your GPU can only fit a small
        batch size but model convergence is hindered by that.
    """

    def __init__(self, model):
        super().__init__(model)
        self._grad_accum_cache = {}

    # MARK: Callback Overrides
    def before_epoch(self, current, num_iterations):
        self._grad_accum_cache['num_iterations'] = num_iterations
        self._grad_accum_cache['current_iteration'] = 0

    def after_batch(self):
        if 'current_iteration' in self._grad_accum_cache:
            self._grad_accum_cache['current_iteration'] += 1

    def after_epoch(self, metrics):
        self._grad_accum_cache = {}
    
    # MARK: Private Methods
    def train_batch(self, data, optimizer=None, loss=None,
                    gradient_accumulation_steps=1):
        if isinstance(optimizer, str):
            if optimizer not in self._cache:
                self._cache[optimizer] = self.optimizer(optimizer)
            optimizer = self._cache[optimizer]

        it = self._grad_accum_cache['current_iteration']
        if it == 0:
            optimizer.zero_grad()

        loss_func = loss

        x, y_true = data
        y_pred = self._forward(x)
        loss = loss_func(y_pred, y_true) / gradient_accumulation_steps

        loss.backward()

        it_is_last = it == self._grad_accum_cache['num_iterations'] - 1
        if it % gradient_accumulation_steps == 0 or it_is_last:
            optimizer.step()
            optimizer.zero_grad()

        return loss.item()

    def eval_batch(self, data):
        x, y_true = data
        y_pred = self._forward(x)
        return y_pred, y_true



#
#  multiprocessing/test_vectorizer.py
#  bxtorch/tests
#
#  Created by Oliver Borchert on May 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import sys
sys.path.append('../../')

import unittest
import time
import torch
import bxtorch.multiprocessing as xmp

class TestVectorizer(unittest.TestCase):

    def setUp(self):
        self.count = None

    def test_on_main_thread(self):
        tensor = torch.arange(100)
        vectorizer = xmp.Vectorizer(divide_by_two, num_workers=0)
        result = vectorizer.process(tensor)
        self.assertTrue(all(torch.as_tensor(result) == tensor / 2))

    def test_batching(self):
        tensor = torch.arange(100)
        vectorizer = xmp.Vectorizer(divide_by_two, num_workers=4)
        result = vectorizer.process(tensor)
        self.assertTrue(all(torch.as_tensor(result) == tensor / 2))

    def test_iterating(self):
        tensor = torch.arange(100)
        vectorizer = xmp.Vectorizer(divide_by_two, num_workers=4)
        result = vectorizer.process(iter(tensor))
        self.assertTrue(all(torch.as_tensor(result) == tensor / 2))

    def test_worker_init(self):
        tensor = torch.arange(100)
        vectorizer = xmp.Vectorizer(
            times, worker_init=factor_from_rank, num_workers=2
        )
        result = vectorizer.process(tensor)
        expected = tensor
        expected[:50] *= 2
        expected[50:] *= 4
        self.assertTrue(all(torch.as_tensor(result) == expected))

    def test_callback(self):
        tensor = torch.arange(100)
        self.count = 0
        def callback():
            self.count += 1
        vectorizer = xmp.Vectorizer(
            divide_by_two, callback_func=callback, num_workers=4
        )
        vectorizer.process(tensor)
        self.assertEqual(self.count, 100)

    def test_speedup(self):
        tensor = torch.arange(100)
        seq_vectorizer = xmp.Vectorizer(divide_by_two_slow, num_workers=0)
        par_vectorizer = xmp.Vectorizer(divide_by_two_slow, num_workers=4)
        
        tic = time.time()
        seq_vectorizer.process(tensor)
        seq_time = time.time() - tic

        tic = time.time()
        par_vectorizer.process(tensor)
        par_time = time.time() - tic

        speedup = seq_time / par_time
        self.assertTrue(
            par_time * 3.5 < seq_time, 
            msg=f"Speedup must be at least 3.5 but is {speedup:.2f}."
        )

def divide_by_two(x):
    return x / 2

def factor_from_rank(rank):
    return (rank + 1) * 2

def times(x, f):
    return x * f

def divide_by_two_slow(x):
    time.sleep(0.05)
    return x / 2


if __name__ == '__main__':
    unittest.main()



#
#  estimator.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/04/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from bxtorch.utils.stdlib import cached_property

class Estimator:
    """
    Estimators are meant to be mixins for PyTorch modules. They extend the
    module with three additional methods:

    * `fit(...)`: This function optimizes the parameters of the model given
        some (input and output) data.
    * `evaluate(...)`: This function estimates the performance of the model by
        returning some suitable metric based on some (input and output) data. 
        This metric is usually used in the `fit` method as well (e.g. an 
        appropriate loss).
    * `predict(...)`: This function performs inference based on some (input)
        data. This method is usually tightly coupled with the model's forward
        method, however, opens additional possibilities such as easy GPU 
        support.

    Usually, the module does not implement these method itself, *unless 
    training is tightly coupled with the model parameters*. An example might be 
    a linear regression module. Normally, however, the module is expected to
    provide an `__engine__` class that accepts the module as first parameter
    (or alternatively overwrite the `engine` cached property). This engine 
    class acts as a default engine which is used whenever one of the methods is 
    called. The arguments for the functions therefore depend on the particular 
    engine that is being used.
    """

    # MARK: Initialization
    def __init__(self, *args, **kwargs):
        super(Estimator, self).__init__(*args, **kwargs)

    # MARK: Computed Properties
    @cached_property
    def engine(self):
        return type(self).__engine__(self)
        
    # MARK: Instance Methods
    def fit(self, *args, **kwargs):
        """
        Optimizes the parameters of the model based on input and output data.
        The parameters are passed to the `train` method of the model's engine.
        """
        return self.engine.train(*args, **kwargs)

    def evaluate(self, *args, **kwargs):
        """
        Estimates the performance of the model by returning some metric based
        on input and output data. The parameters are passed to the `evaluate`
        method of the model's engine.
        """
        return self.engine.evaluate(*args, **kwargs)

    def predict(self, *args, **kwargs):
        """
        Performs inference based on some input data. The parameters are passed
        to the `predict` method of the model's engine.
        """
        return self.engine.predict(*args, **kwargs)



#
#  base.py
#  bxtorch
#
#  Created by Oliver Borchert on 09/05/19.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import os
import time
from abc import ABC, abstractmethod
import torch
import torch.nn as nn
import torch.optim as optim
import torch.cuda as cuda
from torch.utils.data import DataLoader, Dataset, IterableDataset
import bxtorch.multiprocessing as xmp
from bxtorch.nn.callbacks import TrainingCallback, PredictionCallback, \
    CallbackException
import bxtorch.nn.utils as xnnu
from bxtorch.utils.torch import gpu_device, to_device
from bxtorch.utils.stdlib import flatten
from .wrappers import History, Evaluation

class BaseEngine(TrainingCallback, PredictionCallback, ABC):
    """
    A base class for training and evaluating models as well as making 
    predictions. Generally, this class should be seen as _binding_ between a 
    model and some data. Models should always be wrapped in an engine, both 
    when being trained, and when performing inference/evaluation. The 
    engine ensures that the model's environment is correct and prevents plenty 
    of pitfalls.
    
    A concrete implementation of this class is tailored to a specific type of 
    data (e.g. independent, identically distributed data samples) and/or model
    types (e.g. GANs).
    """

    # MARK: Initialization
    def __init__(self, model, dataset_class=None, **dataset_kwargs):
        """
        Initializes a new engine for a specified model.

        Parameters:
        -----------
        - model: torch.nn.Module
            The model to train or evaluate.
        - dataset_class: type like torch.utils.data.Dataset, default: None
            The dataset class to use to make predictions. It must accept as
            first parameter a list of raw elements for prediction. In case no
            class is defined, the ``predict`` function expects to receive a
            dataset or data loader instead of raw elements.
        - dataset_kwargs: keyword arguments
            Arguments passed to the dataset upon initialization.
        """
        self.model = model
        self.device = None
        self._cache = {}
        self._iteration = None

    # MARK: Primary Instance Methods
    def train(self, train_data, val_data=None, epochs=20, val_iterations=None,
              eval_every=None, eval_train=False, callbacks=[], metrics={},
              gpu='auto', **kwargs):
        """
        Method for training the model with the supplied parameters.

        Parameters:
        -----------
        - train_data: torch.DataLoader
            A data loader to obtain training data from. The samples yielded by
            the data loader depend on a specific trainer implementation.
        - val_data: torch.DataLoader, default: None
            A data loader to use for validation. If the loader is an infinite
            data loader, ``val_iterations`` must also be given. If not supplied,
            no validation will be performed.
        - epochs: int, default: 20
            The number of epochs to train for. If the given data is an infinite
            data loader, this value defines the number of iterations (and 
            should most probably be increased).
        - val_iterations: int, default: None
            The number of iterations to perform for validation. Must be given
            only if `val_data` is an infinite data loader. Otherwise, the
            parameter is ignored.
        - eval_every: int, default: None
            How many iterations should pass until validation is called again.
            This should only be set if `train_data` is based on an iterable 
            dataset which has infinite length. Note that everything works as if
            `epochs // eval_every` epochs is trained while eval_every can
            be considered the number of mini-batches.
        - eval_train: bool, default: False
            Whether to compute metrics (apart from the loss) for both the
            validation and the train data. If this flag is set to False,
            metrics will only be computed for the validation data.
        - callbacks: list of bxtorch.nn.TrainingCallback or 
                bxtorch.nn.PredictionCallback, default: []
            The callbacks to use for training and inference. The training
            callbacks and prediction callbacks will be filtered automatically.
        - metrics: dict of str -> func, default: {}
            Metrics to compute during evaluation for the validation data (and
            potentially for the training data). The keys for the metrics define
            their name.
        - gpu: str or bool or int or list of int, default: 'auto'
            Governs, whether training and evaluation should be performed on a
            GPU. If set to True, the GPU with the most amount of free memory is
            selected (if there are multiple GPUs). If set to an integer, the GPU
            with the specified index is used. If set to a list of integers, the
            specified GPUs are used to train and evaluate the model und multiple
            GPUs simultaneously. In this case, the batch sizes of the data
            loaders should be adjusted accordingly. If set to a string, the
            only valid value is 'auto'. In this case, all available GPUs are
            used.
        - kwargs: keyword arguments
            Additional keyword arguments dependent on the specific subclass.

        Returns:
        --------
        - bxtorch.nn.History
            A history object summarizing stats from the training. It contains
            as properties the development of the loss as ``train_loss`` (and
            potentially ``val_loss``, if ``val_data`` is supplied and the 
            keyword arguments include a function called ``loss``). If
            additional metrics are supplied, there will be a property 
            ``val_<metric>`` for each metric, and potentially ``train_<metric>``
            if ``eval_train`` is set to True.
        """
        # 1) Setup
        exception = None
        tic = time.time()

        if eval_every is not None:
            # Here, epochs are considered iterations
            epochs = epochs // eval_every

        # 1.1) Callbacks
        train_callbacks = [
            c for c in callbacks if isinstance(c, TrainingCallback)
        ]
        prediction_callbacks = [
            c for c in callbacks if isinstance(c, PredictionCallback)
        ]

        self._exec_callbacks(
            train_callbacks, 'before_training', self.model, epochs
        )

        # 1.2) Metrics
        metric_history = []
        if 'loss' in kwargs:
            val_metrics = {**metrics, **{'loss': kwargs['loss']}}
        else:
            val_metrics = metrics

        # 1.3) Data loading
        if eval_every is not None:
            train_iterator = iter(train_data)

        # 1.4) GPU support
        if gpu == 'auto':
            if cuda.device_count() == 0:
                gpu = False
            else:
                gpu = list(range(cuda.device_count))
        self._setup_device(gpu)
        self.model.to(self.device)
        
        # 2) Train for number of epochs
        for current_epoch in range(epochs):
            # 2.1) Prepare
            batch_iterations = eval_every or len(train_data)
                
            try:
                self._exec_callbacks(
                    train_callbacks, 'before_epoch', current_epoch, 
                    batch_iterations
                )
            except CallbackException as e:
                exception = e
                break

            # 2.2) Train
            self.model.train()

            train_batch_weights = []
            train_losses = []

            if eval_every is not None:
                # 2.2.1) Iterable dataset
                for _ in range(eval_every):
                    item = next(train_iterator)
                    item = to_device(self.device, item)
                    loss = self.train_batch(item, **kwargs)
                    train_losses.append(loss)
                    self._exec_callbacks(
                        train_callbacks, 'after_batch'
                    )
            else:
                # 2.2.2) Dataset
                for item in train_data:
                    item = to_device(self.device, item)
                    loss = self.train_batch(item, **kwargs)
                    train_batch_weights.append(len(item))
                    train_losses.append(loss)
                    self._exec_callbacks(
                        train_callbacks, 'after_batch'
                    )

            # 2.3) Validate
            batch_metrics = Evaluation(
                self.collate_train_losses(train_losses),
                train_batch_weights
            )

            if val_data is not None:
                eval_val = self.evaluate(
                    val_data, iterations=val_iterations, metrics=val_metrics,
                    callbacks=prediction_callbacks, gpu=None
                ).with_prefix('val_')
                batch_metrics = Evaluation.merge(batch_metrics, eval_val)

            if eval_train:
                eval_train = self.evaluate(
                    train_data, iterations=val_iterations, metrics=metrics,
                    callbacks=prediction_callbacks, gpu=None
                ).with_prefix('train_')
                batch_metrics = Evaluation.merge(batch_metrics, eval_train)

            batch_metrics = Evaluation.merge(
                batch_metrics, Evaluation({'_timestamp': time.time()})
            )

            metric_history.append({
                'micro_train_loss': train_losses,
                **batch_metrics.to_dict()
            })

            # 2.4) Finish epoch
            try:
                self._exec_callbacks(
                    train_callbacks, 'after_epoch', batch_metrics
                )
            except CallbackException as e:
                exception = e
                break

        # 3) Finish training
        # 3.1) If GPU used
        if gpu is not None:
            self.model.to('cpu', non_blocking=True)
            self.device = None

        # 3.2) Finish callbacks
        self._exec_callbacks(
            train_callbacks, 'after_training'
        )
        if exception is not None:
            print(exception)

        return History(time.time() - tic, metric_history)

    def evaluate(self, data, iterations=None, metrics={}, callbacks=[],
                 gpu='auto'):
        """
        Evaluates the model on the given data and computes the supplied metrics.

        Parameters:
        -----------
        - data: torch.DataLoader
            A data loader to obtain evaluation samples from. The expected 
            samples depend on a specific trainer subclass.
        - iterations: int, default: None
            The number of samples used for evaluating if the given data is an
            infinite data loader.
        - metrics: dict of str -> func, default: {}
            The metrics to evaluate the model for. The keys define the names of
            the metrics when retrieving the evaluated result from the return
            parameter.
        - callbacks: list of bxtorch.nn.PredictionCallback, default: []
            Callbacks to use while computing predictions. Usually, they are 
            used for logging.
        - gpu: str or bool or int or list of int, default: False
            Governs, whether training and evaluation should be performed on a
            GPU. If set to True, the GPU with the most amount of free memory is
            selected (if there are multiple GPUs). If set to an integer, the GPU
            with the specified index is used. If set to a list of integers, the
            specified GPUs are used to train and evaluate the model und multiple
            GPUs simultaneously. In this case, the batch sizes of the data
            loaders should be adjusted accordingly. If set to 'auto', all
            available GPUs are used. In case of `None`, the model will not be
            moved. Only use this option if you know what you are doing.

        Returns:
        --------
        - bxtorch.nn.training.wrappers.Evaluation
            An evaluation object, yielding as properties the metrics with their
            specified names.
        """
        num_predictions = iterations or len(data)
        self._exec_callbacks(
            callbacks, 'before_predictions', self.model, num_predictions
        )

        if gpu is not None:
            if gpu == 'auto':
                if cuda.device_count() > 0:
                    gpu = list(range(cuda.device_count()))
                else:
                    gpu = False
                    
            self._setup_device(gpu)
            self.model.to(self.device)

        self.model.eval()

        predictions = []
        targets = []

        iterator = iter(data)
        for _ in range(num_predictions):
            item = next(iterator)
            item = to_device(self.device, item)

            with torch.no_grad():
                prediction, target = self.eval_batch(item)

            predictions.append(to_device('cpu', prediction))
            targets.append(to_device('cpu', target))

            self._exec_callbacks(
                callbacks, 'after_batch'
            )

        self._exec_callbacks(
            callbacks, 'after_predictions'
        )
        
        predictions = self.collate_predictions(predictions)
        targets = self.collate_targets(targets)

        def process_metric(k, m):
            if isinstance(m, dict):
                return [(f'{k}_{mk}', mm.item()) for mk, mm in m.items()]
            return [(k, m.item())]
        
        result = dict(flatten(
            process_metric(k, f(predictions, targets))
            for k, f in metrics.items()
        ))

        if gpu is not None:
            self.model.to('cpu', non_blocking=True)
            self.device = None

        return Evaluation(result)

    def predict(self, data, iterations=None, callbacks=[], gpu='auto',
                **kwargs):
        """
        Computes predictions for the given samples.

        Parameters:
        -----------
        - data: torch.Dataset or torch.DataLoader
            If a data loader is given, it is used directly, otherwise the
            dataset is wrapped into a data loader. Generally, you should prefer
            passing a dataset since the predict function will then make sure
            that items are sampled sequentially from the dataset, and batches
            are pinned when using a GPU.
        - iterations: int, default: None
            The (maximum) number of samples used for evaluating if the given 
            data is an iterable dataset.
        - callbacks: list of bxtorch.nn.PredictionCallback
            Callbacks which are called as prediction progresses.
        - gpu: str or bool or int or list of int, default: 'auto'
            Whether to use a (specific) GPU or multiple GPUs. If multiple GPUs
            are used, one process per GPU is started to minimize
            synchronization overhead. Make sure that using multiple GPUs makes 
            up for this overhead. If ``False`` is specified, all cores of the 
            computer are used to make predictions in parallel. In the case of
            'auto', all available GPUs are used (if any).
        - kwargs: keyword arguments
            Additional arguments fed directly to the data loader. Includes e.g.
            ``batch_size``.

        Returns:
        --------
        - torch.Tensor
            The predictions made by the model.
        - torch.Tensor or object
            The collated targets if `for_eval` is specified.
        """
        # 1) Set gpu if all is specified
        if gpu == 'auto':
            if cuda.device_count() > 0:
                gpu = list(range(cuda.device_count()))
            else:
                gpu = False

        # 2) Get data loader
        if isinstance(data, DataLoader):
            loader = data
        else:
            assert isinstance(data, Dataset), \
                "Data must be either a data loader or a dataset."

            # kwargs['pin_memory'] = isinstance()
            kwargs['shuffle'] = False
            loader = data.loader(**kwargs)
        
        num_iterations = iterations or len(loader)

        self._exec_callbacks(
            callbacks, 'before_predictions', self.model, num_iterations
        )

        # 3) Make sure that the model is not data parallel, we don't need this
        #    for predicting
        model = self.prediction_model.eval()

        # 4) Now perform predictions
        if (isinstance(gpu, list) and len(gpu) > 1) or gpu == False: 
            # parallel computation
            if isinstance(gpu, list):
                num_workers = len(gpu)
            elif isinstance(gpu, bool):
                num_workers = os.cpu_count()
            else:
                num_workers = 1

            model.share_memory()

            callback = lambda:self._exec_callbacks(callbacks, 'after_batch')
            vectorizer = xmp.Vectorizer(
                _prediction_worker_func, _prediction_worker_init, 
                callback_func=callback, num_workers=num_workers,
                gpu=gpu, model=model
            )

            iterator = iter(loader)
            predictions = vectorizer.process(
                (next(iterator) for _ in range(num_iterations)),
                model, self.process_prediction
            )
        else:
            # sequential computation
            device = gpu_device(gpu[0] if isinstance(gpu, list) else gpu)
            model = model.to(device)

            predictions = []

            iterator = iter(loader)
            for _ in range(num_iterations):
                x = next(iterator)

                out = _prediction_worker_func(
                    x, model, self.process_prediction, device
                )
                predictions.append(out)
                self._exec_callbacks(
                    callbacks, 'after_batch'
                )

        self._exec_callbacks(
            callbacks, 'after_predictions'
        )

        return self.collate_predictions(predictions)

    # MARK: Secondary Instance Methods
    def optimizer(self, optimizer='adam', **kwargs):
        """
        Convenience function to create an optimizer for the model from a string.

        Parameters:
        -----------
        - optimizer: str, default: 'adam'
            The optimizer to use. Acceptable values are all optimizers from
            torch.optim (upper and lowercase does not matter).
        - kwargs: keyword arguments
            Arguments passed to the optimizer during initalization.
        """
        valid_optimizers = [
            'ASGD', 'Adadelta', 'Adagrad', 'Adam', 'Adamax', 'LBFGS',
            'RMSprop', 'Rprop', 'SGD', 'SparseAdam'
        ]
        for opt in valid_optimizers:
            if opt.lower() == optimizer.lower():
                return getattr(optim, opt)(self.model.parameters(), **kwargs)
        raise ValueError(f'Invalid optimizer {optimizer}.')

    # MARK: Overrideable Instance Methods
    @abstractmethod
    def train_batch(self, data, **kwargs):
        """
        Runs a single step in training. If the training data represents an 
        infinite dataset, this equals a single iteration, otherwise a 
        mini-batch.

        Parameters:
        -----------
        - data: object
            The data for the current iteration/mini-batch.

        Returns:
        --------
        - object
            The loss computed for the batch. If the returned value is not float,
            overwrite collate_train_losses`.
        """
        pass

    @abstractmethod
    def eval_batch(self, data, **kwargs):
        """
        Runs a single step for inference. The data is either a mini-batch or a 
        single iteration, depending on the data used for evaluation.

        Parameters:
        -----------
        - data: object
            The data for the current iteration/mini-batch.
        - kwargs: keyword arguments
            Additional arguments dependent on the subclass and passed directly
            from the evaluation method.

        Returns:
        --------
        - torch.Tensor
            The output from the model.
        - object
            The target, i.e. correct output. If this is not a ``torch.Tensor``, 
            the ``collate_targets`` should be overriden.
        """
        pass

    def process_prediction(self, x, out):
        """
        Processes a single batch of predictions based on the input and output.
        The default implementation simply returns the output but in some cases
        it might be desirable to have knowledge of the input to enrich the
        output with some information. The output of this function will be
        passed to the `collate_predictions` function.

        Parameters:
        -----------
        - x: torch.Tensor
            The batch inputted to the model.
        - out: torch.Tensor
            The output of the model.

        Returns:
        --------
        - object
            The return value passed to `collate_predictions`.
        """
        return out

    def collate_predictions(self, predictions):
        """
        Combines the predictions obtained from the `eval_batch` function.
        The default implementation assumes that predictions are tensors and can 
        simply be conatenated.

        Parameters:
        -----------
        - predictions: list of objects
            The predictions.

        Returns:
        --------
        - object
            An object to be used as predicted value for some metric.
        """
        return torch.cat(predictions)

    def collate_targets(self, targets):
        """
        Combines the targets in a way that it can be passed to some metric.
        The default implementation assumes that targets are tensors and simply
        concatenates them.

        Parameters:
        -----------
        - list of object
            The targets.

        Returns:
        --------
        - object
            An object to be used as true target with some metric.
        """
        return torch.cat(targets)

    def collate_train_losses(self, losses):
        """
        Combines the losses obtained from the `train_batch function.
        The default implementation assumes that simple floats are returned.

        Parameters:
        -----------
        - losses: list of object
            The losses returned from `train_batch`.
        
        Returns:
        --------
        - dict of str -> (float or list of float)
            The loss names mapped to their values.
        """
        return {'train_loss': losses}

    @property
    def prediction_model(self):
        """
        Defines the model used for prediction. This is useful, e.g. in the
        setting of a GAN where only the generator is used to make predictions.

        Returns:
        --------
        - torch.nn.Module
            The model to use for predictions.
        """
        if isinstance(self.model, nn.DataParallel):
            return self.model.module
        else:
            return self.model

    # MARK: Private Methods
    def _setup_device(self, gpu):
        if isinstance(gpu, list) and len(gpu) > 1:
            self.model = xnnu.DataParallel(self.model, device_ids=gpu)
            self.device = gpu_device(gpu[0])
        else:
            self.device = gpu_device(gpu[0] if isinstance(gpu, list) else gpu)

    def _exec_callbacks(self, callbacks, func, *args):
        for callback in ([self] + callbacks):
            getattr(callback, func)(*args)

    def _forward(self, x):
        return _forward(self.model, x)


def _prediction_worker_func(item, model, process, device):
    x = to_device(device, item)
    with torch.no_grad():
        out = _forward(model, x)
    return to_device('cpu', process(x, out))


def _prediction_worker_init(rank, gpu, model):
    if isinstance(gpu, list):
        device = gpu_device(gpu[rank])
    else:
        device = gpu_device(gpu)
    model.to(device)
    return device


def _forward(model, x):
    if isinstance(x, (list, tuple)):
        return model(*x)
    elif isinstance(x, dict):
        return model(**x)
    else:
        return model(x)



#
#  nn/modules/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .loss import BCELoss
from .lstm import StackedLSTM, StackedLSTMCell



#
#  nn/functional/test_metrics.py
#  bxtorch/tests
#
#  Created by Oliver Borchert on June 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#


import unittest
import time
import numpy as np
import torch
import sklearn.metrics as m
import bxtorch.nn.functional as X

class TestMetrics(unittest.TestCase):

    def test_roc_auc_score(self):
        for _ in range(100):
            y_pred = np.random.uniform(size=(100,))
            y_true = np.random.choice(2, size=(100,))

            roc_pred = X.roc_auc_score(
                torch.from_numpy(y_pred), torch.from_numpy(y_true)
            )
            roc_true = m.roc_auc_score(y_true, y_pred)

            self.assertAlmostEqual(roc_pred, roc_true)


if __name__ == '__main__':
    unittest.main()
